import json

from django.test import TestCase
from datetime import date

from tests.serializers import TemperatureRecordSerializer
from tests.models import WeatherStation, TemperatureRecord

from rest_framework.test import APIClient
from rest_framework.serializers import ModelSerializer
from datapunt_api.rest import DisplayField

# # fake requests
# See: https://stackoverflow.com/questions/34438290/
# from rest_framework.request import Request
# from rest_framework.test import APIRequestFactory
#
# factory = APIRequestFactory()
# request = factory.get('/')

FORMATS = [
    ('api', 'text/html; charset=utf-8'),
    ('json', 'application/json'),
    ('xml', 'application/xml; charset=utf-8'),
    ('csv', 'text/csv; charset=utf-8'),
]


class TestDisplayFieldSerializer(ModelSerializer):
    _display = DisplayField()

    class Meta:
        model = WeatherStation
        fields = '__all__'


class SerializerTest(TestCase):

    def setUp(self):
        ws = WeatherStation.objects.create(number=260)
        records = [
            {'station': ws, 'date': date(1901, 1, 1), 'temperature': '10.0'},
            {'station': ws, 'date': date(1901, 2, 1), 'temperature': '11.0'},
            {'station': ws, 'date': date(1901, 3, 1), 'temperature': '20.0'},
        ]
        for rs in records:
            TemperatureRecord.objects.create(**rs)

    def valid_response(
            self, url, response,
            content_type='text/html; charset=utf-8'):
        """
        Helper method to check common status/json
        """

        self.assertEqual(
            200, response.status_code, "Wrong response code for {}".format(url)
        )

        self.assertEqual(
            f"{content_type}",
            response["Content-Type"],
            "Wrong Content-Type for {}".format(url),
        )

    def test_active(self):
        self.assertTrue(True)

    def test_cannot_serialize_without_request_context(self):
        qs = TemperatureRecord.objects.all()
        self.assertTrue(qs.count() == 3)

        # We need request in context to build hyperlinks for relations.
        with self.assertRaises(AssertionError):
            TemperatureRecordSerializer(qs, many=True).data

    def test_json_html(self):
        c = APIClient()
        response = c.get('/tests/')
        self.assertEqual(response.status_code, 200)

    def test_options_html(self):
        c = APIClient()
        response = c.options('/tests/')
        self.assertEqual(response.status_code, 200)
        response = c.options('/tests/weatherstation/')
        self.assertEqual(response.status_code, 200)

    def test_formats(self):

        c = APIClient()

        for fmt, encoding in FORMATS:
            url = '/tests/weatherstation/'
            response = c.get(url, {'format': fmt})
            self.valid_response(url, response, content_type=encoding)

    def test_detailed(self):
        c = APIClient()
        url = '/tests/weatherstation/'
        response = c.get(url, {'format': 'json', 'detailed': True})
        self.valid_response(url, response, 'application/json')
        data = response.json()
        self.assertEqual(data['results'][0]['detailed'], 'I am detailed')

    def test_weatherstation_endpoint(self):
        c = APIClient()
        response = c.get('/tests/weatherstation/', {'format': 'json'})
        self.assertEqual(response.status_code, 200)
        payload = json.loads(response.content)

        # check that the basics look good for list endpoint
        self.assertEqual(payload['count'], 1)
        self.assertEqual(len(payload['results']), 1)
        self.assertEqual(
            payload['_links']['self']['href'],
            'http://testserver/tests/weatherstation/?format=json'
        )

        # check that we can follow the link to the detail endpoint
        detail_page = payload['results'][0]['_links']['self']['href']

        response = c.get(detail_page, format='json')
        self.assertEqual(response.status_code, 200)
        payload = json.loads(response.content)
        self.assertEqual(
            payload['_links']['self']['href'],
            detail_page
        )
        self.assertEqual(payload['number'], 260)

    def test_hal_style_pagination(self):
        c = APIClient()
        response = c.get('/tests/weatherstation/', format='json')
        self.assertEqual(response.status_code, 200)
        payload = json.loads(response.content)

        self.assertIn('_links', payload)
        self.assertEqual(
            payload['_links']['self']['href'],
            'http://testserver/tests/weatherstation/'
        )
        self.assertEqual(payload['_links']['next']['href'], None)
        self.assertEqual(payload['_links']['previous']['href'], None)

    def test_display_field(self):
        ws = WeatherStation.objects.get(number__exact=260)
        serializer = TestDisplayFieldSerializer(ws)
        self.assertEquals(
            serializer.data['_display'],
            'DISPLAY FIELD CONTENT'
        )



from rest_framework_csv.renderers import CSVRenderer


class PaginatedCSVRenderer(CSVRenderer):
    results_field = 'results'

    def render(self, data, *args, **kwargs):
        if not isinstance(data, list):
            data = data.get(self.results_field, [])
        return super(PaginatedCSVRenderer, self).render(data, *args, **kwargs)



SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'datapunt_api',
    'rest_framework',
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'tests',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

ROOT_URLCONF = 'tests.base_urls'

STATIC_ROOT = 'static/'
STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
            ],
        },
    },
]

REST_FRAMEWORK = dict(
    PAGE_SIZE=100,

    MAX_PAGINATE_BY=100,
    DEFAULT_PAGINATION_CLASS='datapunt_api.pagination.HALPagination',

    DEFAULT_RENDERER_CLASSES=(
       'rest_framework.renderers.JSONRenderer',
       'datapunt_api.renderers.PaginatedCSVRenderer',
       'rest_framework.renderers.BrowsableAPIRenderer',
       'rest_framework_xml.renderers.XMLRenderer'
    ),

    DEFAULT_FILTER_BACKENDS=(
        'django_filters.rest_framework.DjangoFilterBackend',
    ),

    COERCE_DECIMAL_TO_STRING=False,  # other Datapunt projects set this to True
    UNAUTHENTICATED_USER=None,
)



from datapunt_api.rest import DatapuntViewSet

from tests.models import WeatherStation
from tests.models import TemperatureRecord
from tests.serializers import WeatherStationSerializer
from tests.serializers import WeatherDetailStationSerializer
from tests.serializers import TemperatureRecordSerializer


class WeatherStationViewSet(DatapuntViewSet):
    serializer_class = WeatherStationSerializer
    serializer_detail_class = WeatherDetailStationSerializer
    queryset = WeatherStation.objects.all().order_by('id')


class TemperatureRecordViewSet(DatapuntViewSet):
    serializer_class = TemperatureRecordSerializer
    serializer_detail_class = TemperatureRecordSerializer
    queryset = TemperatureRecord.objects.all().order_by('date')

    # test custom inheritance.
    def list(self, request, *args, **kwargs):
        return super().list(self, request, *args, **kwargs)



from collections import OrderedDict
from rest_framework import pagination, response
from rest_framework.utils.urls import replace_query_param


class HALPagination(pagination.PageNumberPagination):
    """
    Implement HAL-JSON style pagination (standard for Datapunt APIs).
    """
    page_size_query_param = 'page_size'

    def get_paginated_response(self, data):
        self_link = self.request.build_absolute_uri()
        if self_link.endswith(".api"):
            self_link = self_link[:-4]

        if self.page.has_next():
            next_link = replace_query_param(
                self_link, self.page_query_param, self.page.next_page_number())
        else:
            next_link = None

        if self.page.has_previous():
            prev_link = replace_query_param(
                self_link, self.page_query_param,
                self.page.previous_page_number())
        else:
            prev_link = None

        return response.Response(OrderedDict([
            ('_links', OrderedDict([
                ('self', dict(href=self_link)),
                ('next', dict(href=next_link)),
                ('previous', dict(href=prev_link)),
            ])),
            ('count', self.page.paginator.count),
            ('results', data)
        ]))



"""
Serialization classes for Datapunt style Django REST Framework APIs.
"""
from collections import OrderedDict

from rest_framework import serializers
from rest_framework.reverse import reverse
import json


def get_links(view_name, kwargs=None, request=None):
    result = OrderedDict([
        ('self', dict(
            href=reverse(view_name, kwargs=kwargs, request=request)
        ))
    ])

    return result


class DataSetSerializerMixin(object):
    def to_representation(self, obj):
        result = super().to_representation(obj)
        result['dataset'] = self.dataset
        return result


class LinksField(serializers.HyperlinkedIdentityField):

    def to_representation(self, value):
        request = self.context.get('request')

        result = OrderedDict([
            ('self', dict(
                href=self.get_url(value, self.view_name, request, None))
             ),
        ])

        return result


class HALSerializer(serializers.HyperlinkedModelSerializer):
    url_field_name = '_links'
    serializer_url_field = LinksField


# TODO: check that this is needed (vs just using the HALSerializer above).
class SelfLinkSerializerMixin():
    def get__links(self, obj):
        """
        Serialization of _links field for detail view (assumes ModelViewSet).

        Note:
            Used to provide HAL-JSON style self links.
        """
        view = self.context['view']
        model = view.queryset.model
        pk_value = getattr(obj, model._meta.pk.name)

        return {
            'self': {
                'href': view.reverse_action('detail', args=[pk_value])
            }
        }


class RelatedSummaryField(serializers.Field):
    def to_representation(self, value):
        count = value.count()
        model_name = value.model.__name__
        mapping = model_name.lower() + "-list"
        url = reverse(mapping, request=self.context['request'])

        parent_pk = value.instance.pk
        filter_name = list(value.core_filters.keys())[0]

        return dict(
            count=count,
            href="{}?{}={}".format(url, filter_name, parent_pk),
        )


# Note about DisplayField below; setting source to '*' causes the
# whole (model) instance to be passed to the DisplayField See:
# http://www.django-rest-framework.org/api-guide/fields/#source
# Display field then uses the __str__ function on the Django
# model to get a nice string representation that can be presented
# to the user.

class DisplayField(serializers.Field):
    """
    Add a `_display` field, based on Model string representation.
    """
    def __init__(self, *args, **kwargs):
        kwargs['source'] = '*'
        kwargs['read_only'] = True
        super().__init__(*args, **kwargs)

    def to_representation(self, value):
        return str(value)


class MultipleGeometryField(serializers.Field):

    read_only = True

    def get_attribute(self, obj):
        # Checking if point geometry exists. If not returning the
        # regular multipoly geometry
        return obj.geometrie

    def to_representation(self, value):
        # Serialize the GeoField. Value could be either None,
        # Point or MultiPoly
        res = ''
        if value:
            res = json.loads(value.geojson)
        return res



from django.db import models


class WeatherStation(models.Model):
    number = models.IntegerField(unique=True)

    def __str__(self):
        return 'DISPLAY FIELD CONTENT'


class TemperatureRecord(models.Model):
    class Meta:
        unique_together = ('station', 'date')

    station = models.ForeignKey(WeatherStation, on_delete=models.CASCADE)
    date = models.DateField()
    temperature = models.DecimalField(decimal_places=3, max_digits=6)



from datapunt_api.rest import HALSerializer
from rest_framework import serializers

from tests import models


class WeatherStationSerializer(HALSerializer):
    class Meta:
        model = models.WeatherStation
        fields = '__all__'


class WeatherDetailStationSerializer(HALSerializer):

    detailed = serializers.SerializerMethodField()

    class Meta:
        model = models.WeatherStation
        fields = [
            '_links',
            'number',
            'detailed'
        ]

    def get_detailed(self, obj):
        return 'I am detailed'


class TemperatureRecordSerializer(HALSerializer):
    class Meta:
        model = models.TemperatureRecord
        fields = '__all__'



"""
Bounding box methods usefull for Amsterdam.
"""
from math import pi, cos
from django.contrib.gis.geos import Point
from rest_framework.serializers import ValidationError

# Default amstermdam bbox lon, lat, lon, lat
# square around Amsterdam.
BBOX = [52.03560, 4.58565,
        52.48769, 5.31360]


def parse_xyr(value: str) -> (Point, int):
    """
    Parse x, y, radius input.
    """
    try:
        x, y, radius = value.split(',')
    except ValueError:
        raise ValidationError(
            "Locatie must be rdx,rdy,radius(m) or lat,long,radius(m)"
        )

    try:
        # Converting , to . and then to float
        x = float(x)
        y = float(y)
        radius = float(radius)
    except ValueError:
        raise ValidationError(
            "Locatie must be x: float, y: float, r: int"
        )

    # Checking if the given coords are in RD, otherwise converting
    if y > 10:
        point = Point(x, y, srid=28992).transform(4326, clone=True)
    else:
        point = Point(x, y, srid=4326)

    radius = dist_to_deg(radius, x)
    return point, radius


def dist_to_deg(distance, latitude):
    """
    distance = distance in meters
    latitude = latitude in degrees
    at the equator, the distance of one degree is equal in latitude and longitude.
    at higher latitudes, a degree longitude is shorter in length, proportional to cos(latitude)
    http://en.wikipedia.org/wiki/Decimal_degrees
    This function is part of a distance filter where the database 'distance' is in degrees.
    There's no good single-valued answer to this problem.
    The distance/ degree is quite constant N/S around the earth (latitude),
    but varies over a huge range E/W (longitude).
    Split the difference: I'm going to average the the degrees latitude and degrees longitude
    corresponding to the given distance. At high latitudes, this will be too short N/S
    and too long E/W. It splits the errors between the two axes.
    Errors are < 25 percent for latitudes < 60 degrees N/S.
    """  # noqa
    #   d * (180 / pi) / earthRadius   ==> degrees longitude
    #   (degrees longitude) / cos(latitude)  ==> degrees latitude

    lat = latitude if latitude >= 0 else -1 * latitude
    rad2deg = 180 / pi
    earthRadius = 6378160.0
    latitudeCorrection = 0.5 * (1 + cos(lat * pi / 180))
    return (distance / (earthRadius * latitudeCorrection) * rad2deg)


def determine_bbox(request):
    """
    Create a bounding box if it is given with the request.
    """

    err = "invalid bbox given"

    if 'bbox' not in request.query_params:
        # set default value
        return BBOX, None

    bboxp = request.query_params['bbox']
    bbox, err = valid_bbox(bboxp)

    if err:
        return None, err

    return bbox, err


def valid_bbox(bboxp):
    """
    Check if bbox is a valid bounding box
    """
    bbox = bboxp.split(',')
    err = None

    # check if we got 4 parametes
    if not len(bbox) == 4:
        return [], "wrong numer of arguments (lon, lat, lon, lat)"

    # check if we got floats
    try:
        bbox = [float(f) for f in bbox]
    except ValueError:
        return [], "Did not recieve floats"

    # max bbox sizes from mapserver
    # RD  EXTENT      100000    450000   150000 500000
    # WGS             52.03560, 4.58565  52.48769, 5.31360
    lat_min = 52.03560
    lat_max = 52.48769
    lon_min = 4.58565
    lon_max = 5.31360

    # check if coorinates are withing amsterdam
    # lat1, lon1, lat2, lon2 = bbox

    # bbox given by leaflet
    lon1, lat1, lon2, lat2 = bbox

    if not lat_max >= lat1 >= lat_min:
        err = f"lat not within max bbox {lat_max} > {lat1} > {lat_min}"

    if not lat_max >= lat2 >= lat_min:
        err = f"lat not within max bbox {lat_max} > {lat2} > {lat_min}"

    if not lon_max >= lon2 >= lon_min:
        err = f"lon not within max bbox {lon_max} > {lon2} > {lon_min}"

    if not lon_max >= lon1 >= lon_min:
        err = f"lon not within max bbox {lon_max} > {lon1} > {lon_min}"

    # this is how the code expects the bbox
    bbox = [lat1, lon1, lat2, lon2]

    return bbox, err



from rest_framework import viewsets
from rest_framework import renderers
from rest_framework_extensions.mixins import DetailSerializerMixin
from django_filters.rest_framework import DjangoFilterBackend

from rest_framework.settings import api_settings

from rest_framework_xml.renderers import XMLRenderer

from .renderers import PaginatedCSVRenderer
from .pagination import HALPagination
from .serializers import (  # noqa: F401
    DisplayField, DataSetSerializerMixin, get_links,
    HALSerializer, LinksField, RelatedSummaryField,
    SelfLinkSerializerMixin
)

DEFAULT_RENDERERS = [
    renderers.JSONRenderer,
    PaginatedCSVRenderer,
    renderers.BrowsableAPIRenderer,
    XMLRenderer,
]


if api_settings.DEFAULT_RENDERER_CLASSES:
    DEFAULT_RENDERERS = api_settings.DEFAULT_RENDERER_CLASSES


class _DisabledHTMLFilterBackend(DjangoFilterBackend):
    """
    See https://github.com/tomchristie/django-rest-framework/issues/3766
    This prevents DRF from generating the filter dropdowns (which can be HUGE
    in our case)
    """

    def to_html(self, request, queryset, view):
        return ""


def _is_detailed_request(detailed_keyword, request):
    value = request.GET.get(detailed_keyword, False)
    if value and value in [1, '1', True, 'True', 'yes']:
        return True


class DatapuntViewSet(DetailSerializerMixin, viewsets.ReadOnlyModelViewSet):
    """
    ViewSet subclass for use in Datapunt APIs.

    Note:
    - this uses HAL JSON style pagination.
    """
    renderer_classes = DEFAULT_RENDERERS
    pagination_class = HALPagination
    filter_backends = (_DisabledHTMLFilterBackend,)
    # TO restore filter box in your view!
    # filter_backends = (DjangoFilterBackend,)
    detailed_keyword = 'detailed'

    def list(self, request, *args, **kwargs):
        # Checking if a detailed response is required
        if _is_detailed_request(self.detailed_keyword, request):
            self.serializer_class = self.serializer_detail_class
        return super().list(self, request, *args, **kwargs)


class DatapuntViewSetWritable(DetailSerializerMixin, viewsets.ModelViewSet):
    """
    ViewSet subclass for use in Datapunt APIs.

    Note:
    - this uses HAL JSON style pagination.
    """
    renderer_classes = DEFAULT_RENDERERS
    pagination_class = HALPagination
    filter_backends = (_DisabledHTMLFilterBackend,)

    detailed_keyword = 'detailed'

    def list(self, request, *args, **kwargs):
        # Checking if a detailed response is required
        if _is_detailed_request(self.detailed_keyword, request):
            self.serializer_class = self.serializer_detail_class
        return super().list(self, request, *args, **kwargs)
